/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include "heat.h"

#define randa(x,t) 0.0f
#define randb(x,t) (exp(-2 * (t)) * sin(x))
#define randc(y,t) 0.0f
#define randd(y,t) (exp(-2 * (t)) * sin(y))

inline float
f(float x, float y)
{
	return sin(x) * sin(y);
}

/*****************   Initialization of grid partition  ********************/

void 
initgrid(struct heat_ctx *params, __global float *buf,
    int lb, int ub)
{
	int a, b, llb, lub;

	llb = (lb == 0) ? 1 : lb;
	lub = (ub == params->nx) ? params->nx - 1 : ub;

	/* boundary nodes */
	for (a = llb, b = 0; a < lub; a++)
		buf[a + params->ny * b] =
		    randa(params->xu + a * params->dx, 0);

	for (a = llb, b = params->ny - 1; a < lub; a++)
		buf[a + params->ny * b] =
		    randb(params->xu + a * params->dx, 0);

	if (lb == 0)
		for (a = 0, b = 0; b < params->ny; b++)
			buf[a + params->ny * b] =
			    randc(params->yu + b * params->dy, 0);
	if (ub == params->nx)
		for (a = params->nx - 1, b=0; b < params->ny; b++)
			buf[a + params->ny * b] =
			    randd(params->yu + b * params->dy, 0);

	/* inner nodes */
	for (a = llb; a < lub; a++)
		for (b = 1; b < params->ny - 1; b++)
			buf[a + params->ny * b] =
			    f(params->xu + a * params->dx,
			    params->yu + b * params->dy);
}

/***************** Five-Point-Stencil Computation ********************/

void 
compstripe(struct heat_ctx *params,
    __constant float *old, __global float *new, int lb, int ub)
{
	int a, b;

	int llb = lb == 0 ? 1 : lb;
	int lub = ub == params->nx ? params->nx - 1 : ub;

	for (a = llb; a < lub; a++)
		for (b = 1; b < params->ny - 1; b++) {
			unsigned pos = a + params->ny * b;
			new[pos] =
			    params->dtdxsq * (old[pos + 1] - 2 * old[pos] +
				old[pos - 1]) +
			    params->dtdysq * (old[pos + params->ny] -
				2 * old[pos] + old[pos - params->ny]) +
				old[pos];
		}

	for (a = llb, b = params->ny - 1; a < lub; a++)
		new[a + params->ny * b] =
		    randb(params->xu + a * params->dx, params->t);

	for (a = llb, b = 0; a < lub; a++)
		new[a + params->ny * b] =
		    randa(params->xu + a * params->dx, params->t);

	if (lb == 0)
		for (a = 0, b = 0; b < params->ny; b++)
			new[a + params->ny * b] =
			    randc(params->yu + b * params->dy, params->t);
	if (ub == params->nx)
		for (a = params->nx - 1, b = 0; b < params->ny; b++)
			new[a + params->ny * b] =
			    randd(params->yu + b * params->dy, params->t);
}

__kernel void
heat(
    /* params */
    struct heat_ctx params,
    unsigned mode,
    __constant float *old,
    __global float *new,

    unsigned max_tid,

    /* input args */
    __constant int *in_lb,
    __constant int *in_ub,

    __global unsigned *result,
    __global unsigned *spawn,

    /* spawn args */
    __global int *spawn_lb,
    __global int *spawn_ub)
{
	unsigned tid = get_global_id(0);
	if (tid > max_tid) return;
	int lb = in_lb[tid];
	int ub = in_ub[tid];

	if (ub - lb > params.leafmaxcol) {
		spawn[tid] = true;
		spawn_lb += tid * 2;
		spawn_ub += tid * 2;

		spawn_lb[0] = lb;
		spawn_ub[0] = (ub + lb) / 2;

		spawn_lb[1] = spawn_ub[0];
		spawn_ub[1] = ub;

		result[tid] = params.leafmaxcol;
		return;
	}

	spawn[tid] = false;
	result[tid] = 1;
	switch (mode) {
	case INIT:
		initgrid(&params, new, lb, ub);
		break;
	case COMP:
		compstripe(&params, old, new, lb, ub);
		break;
	}
}

/* vim: set filetype=c : */
