/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

/*
 * Heat diffusion (Jacobi-type iteration)
 *
 * Usage: see function usage();
 * 
 * Volker Strumpen, Boston                                 August 1996
 *
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

//#include <cilk.h>
//#include <cilk-lib.h>

#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>

#include <cassert>
#include <cerrno>
#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <string>

#include <iomanip>

#include "HRTimer.h"
#include "openclk.hpp"

#include "heat.h"

/* Define ERROR_SUMMARY if you want to check your numerical results */
/*#define ERROR_SUMMARY*/

namespace {

const char *prog;

class Heat : public Alg_spec<cl_uint> {
public:
	Heat(const struct heat_ctx *ctx, cl_uint mode,
	    const cl_float *old, cl_float *new_)
	{
		size_t sz = sizeof(cl_float) * ctx->nx * ctx->ny;

		add_param_generic(*ctx);

		add_param(mode);

		add_param(const_cast<cl_float *>(old), sz,
		    Argument_buffer::read);

		add_param(new_, sz, Argument_buffer::write);

		add_arg(static_cast<cl_int>(0));
		add_arg(static_cast<cl_int>(ctx->nx));
	}

protected:
	virtual std::string name() const
	{	return "heat"; }

	virtual void reduce(const cl_uint *child_results, size_t len,
	    cl_uint *result)
	{
		cl_uint sum = 0;
		for (size_t i = 0; i < len; i++)
			sum += child_results[i];
		*result = sum;
	}

	virtual unsigned sub_tasks() const
	{	return 2; }
};

inline cl_float
solu(cl_float x, cl_float y, cl_float t)
{
	return exp(-2 * t) * sin(x) * sin(y);
}

int nx, ny, nt;
cl_float xu, xo, yu, yo, tu, to;
cl_float dx, dy, dt;

cl_float dtdxsq, dtdysq;
cl_float t;

int leafmaxcol;

}

/***************** Decomposition of 2D grids in stripes ********************/

int 
heat(Openclk &openclk, struct heat_ctx *params)
{
  cl_float *old, *new_;
#ifdef ERROR_SUMMARY  
  cl_float tmp, **mat;
  cl_float mae = 0.0;
  cl_float mre = 0.0;
  cl_float me = 0.0;
  int a, b;
#endif
  
  /* Memory Allocation */
  size_t count = params->nx * params->ny;
  old = new cl_float[count];
  new_ = new cl_float[count];

#if 0
  uint8_t *b = reinterpret_cast<uint8_t *>(params);
  std::cerr << std::hex << std::setfill('0');
  for (size_t i = 0; i < sizeof(*params); i++) {
	  if (i && i % 4 == 0) std::cerr << ' ';
	  std::cerr << std::setw(2) << (short)b[i];
  }
  std::cerr << '\n' << std::dec << std::setfill(' ');
#endif

  /* Initialization */
  {
	Heat init(params, INIT, old, new_);
	init.run(openclk);
  }
  //std::cerr << "initialized.\n";

  /* Jacobi Iteration (divide x-dimension of 2D grid into stripes) */
  for (int c = 0; c < nt; c++) {
	params->t = tu + c * dt;
	std::swap(old, new_);
	Heat comp(params, COMP, old, new_);
  }

  delete[] old;

#ifdef ERROR_SUMMARY  
  /* Error summary computation: Not parallelized! */
  mat = new_;
  printf("\n Error summary of last time frame comparing with exact solution:");
  for (a=0; a<params->nx; a++)
    for (b=0; b<params->ny; b++) {
      tmp = fabs(mat[a][b] - solu(params->xu + a * params->dx, params->yu + b * params->dy, params->to));
      if (tmp > mae)
	mae = tmp;
    }
  printf("\n   Local maximal absolute error  %10e ", mae);
  
  for (a=0; a<params->nx; a++)
    for (b=0; b<params->ny; b++) {
      tmp = fabs(mat[a][b] - solu(params->xu + a * params->dx, params->yu + b * params->dy, params->to));
      if (mat[a][b] != 0.0)
	tmp = tmp / mat[a][b];
      if (tmp > mre)
	mre = tmp;
    }
  printf("\n   Local maximal relative error  %10e %s ", mre * 100, "%");
  
  me = 0.0;
  for (a=0; a<params->nx; a++)
    for (b=0; b<params->ny; b++) {
      me += fabs(mat[a][b] - solu(params->xu + a * params->dx, params->yu + b * params->dy, params->to));
    }
  me = me / (params->nx * params->ny);
  printf("\n   Global Mean absolute error    %10e\n\n", me);
#endif

  delete[] new_;
  return 0;
}

int
usage(void)
{
	std::cerr << "Usage: " << prog << " [<options>]\n"
"\n"
"This program uses a Jacobi-type iteration to solve a finite-difference\n"
"approximation of parabolic partial differential equations that models\n"
"for example the heat diffusion problem.\n\n"

"   -b (short|medium|long)  the benchmark to run (overrides most options)\n"
"   -h        show help\n"
"   -f PATH   input parameters from file\n"
"\n"
"   --gran #  granularity (columns per partition)  default: 10\n"
"   --nx #    total number of columns              default: 4096\n"
"   --ny #    total number of rows                 default: 512\n"
"   --nt #    total time steps                     default: 100\n";
	/*
"   -xu #    lower x coordinate default: 0.0\n"
"   -xo #    upper x coordinate default: 1.570796326794896558\n"
"   -yu #    lower y coordinate default: 0.0\n"
"   -yo #    upper y coordinate default: 1.570796326794896558\n"
"   -tu #    start time         default: 0.0\n"
"   -to #    end time           default: 0.0000001\n"
"   -f filename    parameter file for nx, ny, ...\n"
	 */
	return 1;
}

void 
read_heatparams(const char *filefn)
{
  FILE *f;
  int l;
  
  if ((f = fopen(filefn, "r")) == NULL) {
    printf("\n Can't open %s\n", filefn);
    exit(0);
  }
  l = fscanf(f, "%d %d %d %f %f %f %f %f %f",
	     &nx, &ny, &nt, &xu, &xo, &yu, &yo, &tu, &to);
  if (l != 9)
    printf("\n Warning: fscanf errno %d", errno);
  fclose(f);
  
}

const char *specifiers[] = { "-g", "-nx", "-ny", "-nt", "-xu", "-xo", "-yu", "-yo", "-tu", "-to", "-f", "-benchmark", "-h", 0};
//int opt_types[] = {INTARG, INTARG, INTARG, INTARG, DOUBLEARG, DOUBLEARG, DOUBLEARG, DOUBLEARG, DOUBLEARG, DOUBLEARG, STRINGARG, BENCHMARK, BOOLARG, 0 };

int
main(int argc, char *argv[])
{
  int ret, benchmark;
  std::string filename;

  prog = *argv;

  nx = 512;
  ny = 512;
  nt = 100;
  xu = 0.0;
  xo = 1.570796326794896558;
  yu = 0.0;
  yo = 1.570796326794896558;
  tu = 0.0;
  to = 0.0000001;
  leafmaxcol = 10;
  filename[0]=0;

  if (argc == 1) {
	usage();
	return 1;
  }
  enum longopt_vals { OPT_BEGIN=0x7f,
	  OPT_GRANULARITY, OPT_XSTEPS, OPT_YSTEPS, OPT_TIMESTEPS
  };
  const struct option longopts[] = {
	{ "gran", true, 0, OPT_GRANULARITY },
	{ "nx", true, 0, OPT_XSTEPS },
	{ "ny", true, 0, OPT_YSTEPS },
	{ "nt", true, 0, OPT_TIMESTEPS },
	{ 0, 0, 0, 0 }
  };
  opterr = 0;
  benchmark = 0;
  for (;;) {
	int opt = getopt_long(argc, argv, "-b:f:h", longopts, 0);
	if (opt == -1) break;
	switch (opt) {
	case 1:
		std::cerr << prog << ": no positional options\n";
		return 1;
	case 'b':
		if (!strcmp(optarg, "short"))
			benchmark = 1;
		else if (!strcmp(optarg, "medium"))
			benchmark = 2;
		else if (!strcmp(optarg, "long"))
			benchmark = 3;
		else {
			std::cerr << prog << ": unrecognized benchmark\n";
			return 1;
		}
		break;
	case 'f':
		filename = optarg;
		break;
	case 'h':
		usage();
		return 0;
	case OPT_GRANULARITY:
		leafmaxcol = atoi(optarg);
		break;
	case OPT_XSTEPS:
		nx = atoi(optarg);
		break;
	case OPT_YSTEPS:
		ny = atoi(optarg);
		break;
	case OPT_TIMESTEPS:
		nt = atoi(optarg);
		break;
	case '?':
		std::cerr << prog << ": unrecognized option `-"
		    << static_cast<char>(opt) << "'\n";
		return 1;
	case ':':
		std::cerr << prog << ": missing argument for option `-"
		    << static_cast<char>(opt) << "'\n";
		return 1;
	default:
		assert(0);
	}
  }
  if (!benchmark) {
	std::cerr << prog << ": benchmark not selected\n";
	return 1;
  }

  //get_options(argc, argv, specifiers, opt_types, &leafmaxcol, &nx, &ny, &nt, &xu, &xo, &yu, &yo, &tu, &to, filename, &benchmark, &help);

  if (benchmark) {
    switch (benchmark) {
    case 1:      /* short benchmark options -- a little work*/
      nx = 512;
      ny = 512;
      nt = 1;
      xu = 0.0;
      xo = 1.570796326794896558;
      yu = 0.0;
      yo = 1.570796326794896558;
      tu = 0.0;
      to = 0.0000001;
      leafmaxcol = 10;
      filename[0]=0;
      break;
    case 2:      /* standard benchmark options*/
      nx = 4096;
      ny = 512;
      nt = 40;
      xu = 0.0;
      xo = 1.570796326794896558;
      yu = 0.0;
      yo = 1.570796326794896558;
      tu = 0.0;
      to = 0.0000001;
      leafmaxcol = 10;
      filename[0]=0;
      break;
    case 3:      /* long benchmark options -- a lot of work*/
      nx = 4096;
      ny = 1024;
      nt = 100;
      xu = 0.0;
      xo = 1.570796326794896558;
      yu = 0.0;
      yo = 1.570796326794896558;
      tu = 0.0;
      to = 0.0000001;
      leafmaxcol = 1;
      filename[0]=0;
      break;
    }
  }

  init_timer();

  if (!filename.empty()) read_heatparams(filename.c_str());

  dx = (xo - xu) / (nx - 1);
  dy = (yo - yu) / (ny - 1);
  dt = (to - tu) / nt;	/* nt effective time steps! */
  
  dtdxsq = dt / (dx * dx);
  dtdysq = dt / (dy * dy);

  struct heat_ctx state;
  state.nx = nx;
  state.ny = ny;
  state.nt = nt;
  state.xu = xu;
  state.xo = xo;
  state.yu = yu;
  state.yo = yo;
  state.tu = tu;
  state.to = to;
  state.dx = dx;
  state.dy = dy;
  state.dt = dt;
  state.dtdxsq = dtdxsq;
  state.dtdysq = dtdysq;
  state.t = t;
  state.leafmaxcol = leafmaxcol;

  // initialize Openclk
  const char *source_files[] = {
	"heat.cl",
  };
  size_t count = sizeof(source_files) / sizeof(*source_files);
  Openclk openclk(source_files, source_files + count);
  
  std::cout << "start\n";
  double begin = get_time_ns();
  ret = heat(openclk, &state);
  double end = get_time_ns();

  printf("\nCilk Example: heat\n");
  //printf("	      running on %d processor%s\n\n", Cilk_active_size, Cilk_active_size > 1 ? "s" : "");
  printf("\n   dx = %f", dx);
  printf("\n   dy = %f", dy);
  printf("\n   dt = %f", dt);

  printf("\n\n Stability Value for explicit method must be > 0:  %f\n\n",
	 0.5 - (dt / (dx * dx) + dt / (dy * dy)));
  printf("Options: granularity = %d\n", leafmaxcol);
  printf("         nx          = %d\n", nx);
  printf("         ny          = %d\n", ny);
  printf("         nt          = %d\n", nt);

  //printf("Running time  = %4f s\n", Cilk_wall_time_to_sec(tm_elapsed));
  //printf("Work          = %4f s\n", Cilk_time_to_sec(wk_elapsed));
  //printf("Critical path = %4f s\n\n", Cilk_time_to_sec(cp_elapsed));
  printf("Time: %4lf s\n", (end - begin) / 1000000000.0);

  return 0;
}
