/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

/*
 * Heat diffusion (Jacobi-type iteration)
 *
 * Usage: see function usage();
 * 
 * Volker Strumpen, Boston                                 August 1996
 *
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

//#include <cilk.h>
//#include <cilk-lib.h>

#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>

#include <cassert>
#include <cerrno>
#include <cmath>
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <string>

#include "HRTimer.h"

const char *prog;

/* Define ERROR_SUMMARY if you want to check your numerical results */
/*#define ERROR_SUMMARY*/

#define f(x,y)     (sin(x)*sin(y))
#define randa(x,t) (0.0)
#define randb(x,t) (exp(-2*(t))*sin(x))
#define randc(y,t) (0.0)
#define randd(y,t) (exp(-2*(t))*sin(y))
#define solu(x,y,t) (exp(-2*(t))*sin(x)*sin(y))

int nx, ny, nt;
double xu, xo, yu, yo, tu, to;
double dx, dy, dt;

double dtdxsq, dtdysq;
double t;

int leafmaxcol;

//Cilk_time tm_begin, tm_elapsed;
//Cilk_time wk_begin, wk_elapsed;
//Cilk_time cp_begin, cp_elapsed;


/*****************   Allocation of grid partition  ********************/

void 
allcgrid(double **new_, double **old, int lb, int ub)
{
  int j;
  double **rne, **rol;

  for (j=lb, rol=old+lb, rne=new_+lb; j < ub; j++, rol++, rne++) {
    *rol = new double[ny];
    *rne = new double[ny];
  }
}

/*****************   Initialization of grid partition  ********************/

void 
initgrid(double **old, int lb, int ub)
{
  int a, b, llb, lub;

  llb = (lb == 0) ? 1 : lb;
  lub = (ub == nx) ? nx - 1 : ub;
  
  for (a=llb, b=0; a < lub; a++)		/* boundary nodes */
    old[a][b] = randa(xu + a * dx, 0);
  
  for (a=llb, b=ny-1; a < lub; a++)
    old[a][b] = randb(xu + a * dx, 0);
  
  if (lb == 0) {
    for (a=0, b=0; b < ny; b++)
      old[a][b] = randc(yu + b * dy, 0);
  }
  if (ub == nx) {
    for (a=nx-1, b=0; b < ny; b++)
      old[a][b] = randd(yu + b * dy, 0);
  }
  for (a=llb; a < lub; a++) {	/* inner nodes */
    for (b=1; b < ny-1; b++) {
      old[a][b] = f(xu + a * dx, yu + b * dy);
    }
  }
}


/***************** Five-Point-Stencil Computation ********************/

void 
compstripe(register double **new_, register double **old, int lb, int ub)
{
  register int a, b, llb, lub;
  
  llb = (lb == 0) ? 1 : lb;
  lub = (ub == nx) ? nx - 1 : ub;
  
  for (a=llb; a < lub; a++) {
    for (b=1; b < ny-1; b++) {
      new_[a][b] =   dtdxsq * (old[a+1][b] - 2 * old[a][b] + old[a-1][b])
	          + dtdysq * (old[a][b+1] - 2 * old[a][b] + old[a][b-1])
  	          + old[a][b];
    }
  }
  
  for (a=llb, b=ny-1; a < lub; a++)
    new_[a][b] = randb(xu + a * dx, t);
  
  for (a=llb, b=0; a < lub; a++)
    new_[a][b] = randa(xu + a * dx, t);
  
  if (lb == 0) {
    for (a=0, b=0; b < ny; b++)
      new_[a][b] = randc(yu + b * dy, t);
  }
  if (ub == nx) {
    for (a=nx-1, b=0; b < ny; b++)
      new_[a][b] = randd(yu + b * dy, t);
  }
}


/***************** Decomposition of 2D grids in stripes ********************/

#define ALLC       0
#define INIT       1
#define COMP       2

int 
divide(int lb, int ub, double **new_, double **old, int mode, int timestep)
{
  int l, r;

  if (ub - lb > leafmaxcol) {
    l = /*spawn*/ divide(lb, (ub + lb) / 2, new_, old, mode, timestep);
    r = /*spawn*/ divide((ub + lb) / 2, ub, new_, old, mode, timestep);
    /*sync*/;
    return (l + r);
  } else {
    switch (mode) {
    case COMP:
      if (timestep % 2)
	compstripe(new_, old, lb, ub);
      else
	compstripe(old, new_, lb, ub);
      return 1;
      
    case ALLC:
      allcgrid(new_, old, lb, ub);
      return 1;
      
    case INIT:
      initgrid(old, lb, ub);
      return 1;
    }
  }
  return 0;
}


int 
heat(void)
{
  double **old, **new_;
  int  c, l;
#ifdef ERROR_SUMMARY  
  double tmp, **mat;
  double mae = 0.0;
  double mre = 0.0;
  double me = 0.0;
  int a, b;
#endif
  
  /* Memory Allocation */
  old = new double *[nx];
  new_ = new double *[nx];
  l = /*spawn*/ divide(0, nx, new_, old, ALLC, 0);
  /*sync*/;
  /* 
   * Sequential allocation might be faster than parallel!
   * Substitute the following call for the preceding divide:
   *
   * allcgrid(new_, old, 0, nx);
   */

  /* Initialization */
  l = /*spawn*/ divide(0, nx, new_, old, INIT, 0);
  /*sync*/;

  /* Jacobi Iteration (divide x-dimension of 2D grid into stripes) */
  /* Timing. "Start" timers */
  //cp_begin = Cilk_user_critical_path;
  //wk_begin = Cilk_user_work;
  //tm_begin = Cilk_get_wall_time();

  for (c = 1; c <= nt; c++) {
    t = tu + c * dt;
    
    l = /*spawn*/ divide(0, nx, new_, old, COMP, c);
    /*sync*/;
  }

  /* Timing. "Stop" timers */
  //tm_elapsed = Cilk_get_wall_time() - tm_begin;
  //wk_elapsed = Cilk_user_work - wk_begin;
  //cp_elapsed = Cilk_user_critical_path - cp_begin;



#ifdef ERROR_SUMMARY  
  /* Error summary computation: Not parallelized! */
  mat = (c % 2) ? old : new_;
  printf("\n Error summary of last time frame comparing with exact solution:");
  for (a=0; a<nx; a++)
    for (b=0; b<ny; b++) {
      tmp = fabs(mat[a][b] - solu(xu + a * dx, yu + b * dy, to));
      if (tmp > mae)
	mae = tmp;
    }
  printf("\n   Local maximal absolute error  %10e ", mae);
  
  for (a=0; a<nx; a++)
    for (b=0; b<ny; b++) {
      tmp = fabs(mat[a][b] - solu(xu + a * dx, yu + b * dy, to));
      if (mat[a][b] != 0.0)
	tmp = tmp / mat[a][b];
      if (tmp > mre)
	mre = tmp;
    }
  printf("\n   Local maximal relative error  %10e %s ", mre * 100, "%");
  
  me = 0.0;
  for (a=0; a<nx; a++)
    for (b=0; b<ny; b++) {
      me += fabs(mat[a][b] - solu(xu + a * dx, yu + b * dy, to));
    }
  me = me / (nx * ny);
  printf("\n   Global Mean absolute error    %10e\n\n", me);
#endif
  return 0;
}

int
usage(void)
{
	std::cerr << "Usage: " << prog << " [<options>]\n"
"\n"
"This program uses a Jacobi-type iteration to solve a finite-difference\n"
"approximation of parabolic partial differential equations that models\n"
"for example the heat diffusion problem.\n\n"

"   -b (short|medium|long)  the benchmark to run (overrides most options)\n"
"   -h        show help\n"
"   -f PATH   input parameters from file\n"
"\n"
"   --gran #  granularity (columns per partition)  default: 10\n"
"   --nx #    total number of columns              default: 4096\n"
"   --ny #    total number of rows                 default: 512\n"
"   --nt #    total time steps                     default: 100\n";
	/*
"   -xu #    lower x coordinate default: 0.0\n"
"   -xo #    upper x coordinate default: 1.570796326794896558\n"
"   -yu #    lower y coordinate default: 0.0\n"
"   -yo #    upper y coordinate default: 1.570796326794896558\n"
"   -tu #    start time         default: 0.0\n"
"   -to #    end time           default: 0.0000001\n"
"   -f filename    parameter file for nx, ny, ...\n"
	 */
	return 1;
}

void 
read_heatparams(const char *filefn)
{
  FILE *f;
  int l;
  
  if ((f = fopen(filefn, "r")) == NULL) {
    printf("\n Can't open %s\n", filefn);
    exit(0);
  }
  l = fscanf(f, "%d %d %d %lf %lf %lf %lf %lf %lf",
	     &nx, &ny, &nt, &xu, &xo, &yu, &yo, &tu, &to);
  if (l != 9)
    printf("\n Warning: fscanf errno %d", errno);
  fclose(f);
  
}

const char *specifiers[] = { "-g", "-nx", "-ny", "-nt", "-xu", "-xo", "-yu", "-yo", "-tu", "-to", "-f", "-benchmark", "-h", 0};
//int opt_types[] = {INTARG, INTARG, INTARG, INTARG, DOUBLEARG, DOUBLEARG, DOUBLEARG, DOUBLEARG, DOUBLEARG, DOUBLEARG, STRINGARG, BENCHMARK, BOOLARG, 0 };

int
main(int argc, char *argv[])
{
  int ret, benchmark;
  std::string filename;

  prog = *argv;

  nx = 512;
  ny = 512;
  nt = 100;
  xu = 0.0;
  xo = 1.570796326794896558;
  yu = 0.0;
  yo = 1.570796326794896558;
  tu = 0.0;
  to = 0.0000001;
  leafmaxcol = 10;
  filename[0]=0;

  if (argc == 1) {
	usage();
	return 1;
  }
  enum longopt_vals { OPT_BEGIN=0x7f,
	  OPT_GRANULARITY, OPT_XSTEPS, OPT_YSTEPS, OPT_TIMESTEPS
  };
  const struct option longopts[] = {
	  { "gran", true, 0, OPT_GRANULARITY },
	  { "nx", true, 0, OPT_XSTEPS },
	  { "ny", true, 0, OPT_YSTEPS },
	  { "nt", true, 0, OPT_TIMESTEPS },
	  { 0, 0, 0, 0 }
  };
  opterr = 0;
  benchmark = 0;
  for (;;) {
	int opt = getopt_long(argc, argv, "-b:f:h", longopts, 0);
	if (opt == -1) break;
	switch (opt) {
	case 1:
		std::cerr << prog << ": no positional options\n";
		return 1;
	case 'b':
		if (!strcmp(optarg, "short"))
			benchmark = 1;
		else if (!strcmp(optarg, "medium"))
			benchmark = 2;
		else if (!strcmp(optarg, "long"))
			benchmark = 3;
		else {
			std::cerr << prog << ": unrecognized benchmark\n";
			return 1;
		}
		break;
	case 'f':
		filename = optarg;
		break;
	case 'h':
		usage();
		return 0;
	case OPT_GRANULARITY:
		leafmaxcol = atoi(optarg);
		break;
	case OPT_XSTEPS:
		nx = atoi(optarg);
		break;
	case OPT_YSTEPS:
		ny = atoi(optarg);
		break;
	case OPT_TIMESTEPS:
		nt = atoi(optarg);
		break;
	case '?':
		std::cerr << prog << ": unrecognized option `-"
		    << static_cast<char>(opt) << "'\n";
		return 1;
	case ':':
		std::cerr << prog << ": missing argument for option `-"
		    << static_cast<char>(opt) << "'\n";
		return 1;
	default:
		assert(0);
	}
  }
  if (!benchmark) {
	std::cerr << prog << ": benchmark not selected\n";
	return 1;
  }

  //get_options(argc, argv, specifiers, opt_types, &leafmaxcol, &nx, &ny, &nt, &xu, &xo, &yu, &yo, &tu, &to, filename, &benchmark, &help);

  if (benchmark) {
    switch (benchmark) {
    case 1:      /* short benchmark options -- a little work*/
      nx = 512;
      ny = 512;
      nt = 1;
      xu = 0.0;
      xo = 1.570796326794896558;
      yu = 0.0;
      yo = 1.570796326794896558;
      tu = 0.0;
      to = 0.0000001;
      leafmaxcol = 10;
      filename[0]=0;
      break;
    case 2:      /* standard benchmark options*/
      nx = 4096;
      ny = 512;
      nt = 40;
      xu = 0.0;
      xo = 1.570796326794896558;
      yu = 0.0;
      yo = 1.570796326794896558;
      tu = 0.0;
      to = 0.0000001;
      leafmaxcol = 10;
      filename[0]=0;
      break;
    case 3:      /* long benchmark options -- a lot of work*/
      nx = 4096;
      ny = 1024;
      nt = 100;
      xu = 0.0;
      xo = 1.570796326794896558;
      yu = 0.0;
      yo = 1.570796326794896558;
      tu = 0.0;
      to = 0.0000001;
      leafmaxcol = 1;
      filename[0]=0;
      break;
    }
  }

  init_timer();

  if (!filename.empty()) read_heatparams(filename.c_str());

  dx = (xo - xu) / (nx - 1);
  dy = (yo - yu) / (ny - 1);
  dt = (to - tu) / nt;	/* nt effective time steps! */
  
  dtdxsq = dt / (dx * dx);
  dtdysq = dt / (dy * dy);
  
  double begin = get_time_ns();
  ret = heat();
  double end = get_time_ns();

  printf("\nCilk Example: heat\n");
  //printf("	      running on %d processor%s\n\n", Cilk_active_size, Cilk_active_size > 1 ? "s" : "");
  printf("\n   dx = %f", dx);
  printf("\n   dy = %f", dy);
  printf("\n   dt = %f", dt);

  printf("\n\n Stability Value for explicit method must be > 0:  %f\n\n",
	 0.5 - (dt / (dx * dx) + dt / (dy * dy)));
  printf("Options: granularity = %d\n", leafmaxcol);
  printf("         nx          = %d\n", nx);
  printf("         ny          = %d\n", ny);
  printf("         nt          = %d\n", nt);

  //printf("Running time  = %4f s\n", Cilk_wall_time_to_sec(tm_elapsed));
  //printf("Work          = %4f s\n", Cilk_time_to_sec(wk_elapsed));
  //printf("Critical path = %4f s\n\n", Cilk_time_to_sec(cp_elapsed));
  printf("Time: %4lf s\n", (end - begin) / 1000000000.0);

  return 0;
}
