/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

/* From the cilk people*/

/*
 * Cilk program to solve the 0-1 knapsack problem using a branch-and-bound
 * technique. 
 *
 * Author: Matteo Frigo
 */
static const char *ident __attribute__((__unused__))
     = "$HeadURL: https://bradley.csail.mit.edu/svn/repos/cilk/5.4.3/examples/knapsack.cilk $ $LastChangedBy: sukhaj $ $Rev: 517 $ $Date: 2003-10-27 10:05:37 -0500 (Mon, 27 Oct 2003) $";
/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <climits>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <limits>

#include "HRTimer.h"
#include "openclk.hpp"

/* every item in the knapsack has a weight and a value */
#define MAX_ITEMS 256
struct item {
  int value;
  int weight;
};


namespace {

  class Knapsack : public Alg_spec<cl_int> {
    public:
      Knapsack(cl_int *weights, cl_int *values, cl_int capacity, cl_int n,
	cl_int *best_so_far)
      {
        *best_so_far = std::numeric_limits<cl_int>::min();
        add_param(weights, n*sizeof(cl_int), Argument_buffer::read);
        add_param(values, n*sizeof(cl_int), Argument_buffer::read);
        add_param(best_so_far, sizeof(cl_int), Argument_buffer::read_write);
        add_arg(static_cast<cl_int>(0));
        add_arg(capacity);
        add_arg(n);
        add_arg(static_cast<cl_int>(0));
      }

    protected:
      virtual std::string name() const
      { return "knapsack"; }

      virtual void reduce(const cl_int *child_results, size_t len,
                          cl_int *result)
      {
        cl_int best = std::numeric_limits<cl_int>::min();
        for (size_t i = 0; i < len; i++){
          if (best < child_results[i]){
            best = child_results[i];
          }
        }
	*result = best;
      }

      virtual unsigned sub_tasks() const
      { return 2; }
  };
  
  // wooo!
  int
  knapsack(Openclk &openclk, cl_int *weights, cl_int *values,
    cl_int capacity, cl_int n)
  {
    cl_int best_so_far;
    Knapsack knapsack(weights, values, capacity, n, &best_so_far);
    int res = knapsack.run(openclk);
    std::cerr << "best: " << best_so_far << '\n';
    return res;
  }

} // end anon namespace

int compare(struct item *a, struct item *b)
{
     double c = ((double) a->value / a->weight) -
     ((double) b->value / b->weight);

     if (c > 0)
          return -1;
     if (c < 0)
          return 1;
     return 0;
}

int read_input(const char *filename, struct item *items, int *capacity, int *n)
{
     int i;
     FILE *f;

     if (filename == NULL)
          filename = "\0";
     f = fopen(filename, "r");
     if (f == NULL) {
          fprintf(stderr, "open_input(\"%s\") failed\n", filename);
          return -1;
     }
     /* format of the input: #items capacity value1 weight1 ... */
     assert(fscanf(f, "%d", n) == 1);
     assert(fscanf(f, "%d", capacity) == 1);

     for (i = 0; i < *n; ++i)
          assert(fscanf(f, "%d %d", &items[i].value, &items[i].weight) == 2);

     fclose(f);

     /* sort the items on decreasing order of value/weight */
     /* cilk2c is fascist in dealing with pointers, whence the ugly cast */
     qsort(items, *n, sizeof(struct item),
            (int (*)(const void *, const void *)) compare);

     return 0;
}


int /*cilk_*/main(int argc, char *argv[])
{
     struct item items[MAX_ITEMS];      /* array of items */
     int n, sol, capacity;

     if (read_input("knapsack-example1.input", items, &capacity, &n))
          return 1;

     //due to openClunk limitations, we'll just convert this array of structs
     // into parallel arrays

     cl_int *weights = new cl_int[n];
     cl_int *values = new cl_int[n];

     for (int i=0; i < n; i++){
       weights[i] = items[i].weight;
       values[i] = items[i].value;
     }

     const char *source_files[] = {
       "knapsack.cl",
     };
     size_t count = sizeof(source_files) / sizeof(*source_files);

     // initialize 
     Openclk openclk(source_files, source_files + count);
     
     /* Timing. "Start" timers */
     double start, end;
     init_timer();
     start = get_time_ns();

     sol = knapsack(openclk, weights, values, capacity, n);
//   sync;
     end = get_time_ns();

     delete[] weights;
     delete[] values;

     std::cout << "\nCilk Example: knapsack\n";
     std::cout << "Time: " << (end - start) / 1000000000.0 << " s\n";
     std::cout << "Best value is " << sol << "\n\n";

     return 0;
}
