/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

/* From the cilk people*/

/*
 * Cilk program to solve the 0-1 knapsack problem using a branch-and-bound
 * technique. 
 *
 * Author: Matteo Frigo
 */
static const char *ident __attribute__((__unused__))
     = "$HeadURL: https://bradley.csail.mit.edu/svn/repos/cilk/5.4.3/examples/knapsack.cilk $ $LastChangedBy: sukhaj $ $Rev: 517 $ $Date: 2003-10-27 10:05:37 -0500 (Mon, 27 Oct 2003) $";
/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

// #include <cilk-lib.cilkh>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
// #include <getoptions.h>
#include <string.h>

#include <iostream>
#include "HRTimer.h"

/* every item in the knapsack has a weight and a value */
#define MAX_ITEMS 256

struct item {
     int value;
     int weight;
};

int best_so_far = 0;

int compare(struct item *a, struct item *b)
{
     double c = ((double) a->value / a->weight) -
     ((double) b->value / b->weight);

     if (c > 0)
          return -1;
     if (c < 0)
          return 1;
     return 0;
}

int read_input(const char *filename, struct item *items, int *capacity, int *n)
{
     int i;
     FILE *f;

     if (filename == NULL)
          filename = "\0";
     f = fopen(filename, "r");
     if (f == NULL) {
          fprintf(stderr, "open_input(\"%s\") failed\n", filename);
          return -1;
     }
     /* format of the input: #items capacity value1 weight1 ... */
     fscanf(f, "%d", n);
     fscanf(f, "%d", capacity);

     for (i = 0; i < *n; ++i)
          fscanf(f, "%d %d", &items[i].value, &items[i].weight);

     fclose(f);

     /* sort the items on decreasing order of value/weight */
     /* cilk2c is fascist in dealing with pointers, whence the ugly cast */
     qsort(items, *n, sizeof(struct item),
            (int (*)(const void *, const void *)) compare);

     return 0;
}

/* 
 * return the optimal solution for n items and
 * capacity c. Value so far is v.
 */
/*cilk */int knapsack(int *weights, int *values, int index, int c, int n, int v)
{
     int with, without, best;
     double ub;

     /* base case: full knapsack or no items */
     if (c < 0)
          return INT_MIN;

     if (n == 0 || c == 0)
          return v;             /* feasible solution, with value v */

     ub = (double) v + c * values[index] / weights[index];

     if (ub < best_so_far) {
          /* prune ! */
          return INT_MIN;
     }
     /* 
      * compute the best solution without the current item in the knapsack 
      */
     without = /*spawn */knapsack(weights, values, index+1, c, n - 1, v);

     /* compute the best solution with the current item in the knapsack */
     with = /*spawn */knapsack(weights, values, index+1, c - weights[index], n - 1, v + values[index]);

//      sync;

     best = with > without ? with : without;

     /* 
      * notice the race condition here. The program is still
      * correct, in the sense that the best solution so far
      * is at least best_so_far. Moreover best_so_far gets updated
      * when returning, so eventually it should get the right
      * value. The program is highly non-deterministic.
      */
     if (best > best_so_far)
          best_so_far = best;

     return best;
}

int /*cilk_*/main(int argc, char *argv[])
{
     struct item items[MAX_ITEMS];      /* array of items */
     int n, capacity, sol;

     if (read_input("knapsack-example1.input", items, &capacity, &n))
          return 1;
     
     //due to openClunk limitations, we'll just convert this array of structs
     // into parallel arrays
     
     int *weights = new int[n];
     int *values = new int[n];
     
     for (int i=0; i < n; i++){
       weights[i] = items[i].weight;
       values[i] = items[i].value;
     }

     /* Timing. "Start" timers */
     double start, end;
     init_timer();
     start = get_time_ns();

     sol = /*spawn */knapsack(weights, values, 0, capacity, n, 0);
//      sync;
     end = get_time_ns();

     delete[] weights;
     delete[] values;

     std::cout << "\nCilk Example: knapsack\n";
     std::cout << "Time: " << (end - start) / 1000000000.0 << " s\n";
     std::cout << "Best value is " << sol << "\n\n";

     return 0;
}
