/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <cassert>
#include <fstream>
#include <iostream>

#include "openclk.hpp"

namespace {

cl::Context	create_opencl_context(std::vector<cl::Device> &)
		    throw (Opencl_error);

void		create_opencl_kernels(cl::Context &,
		    const std::vector<cl::Device> &,
		    const std::vector<std::string> &src_paths,
		    std::map<std::string, cl::Kernel> &kernels)
		    throw (Opencl_error);

void		read_file(const std::string &, std::string &);


cl::Context
create_opencl_context(std::vector<cl::Device> &out_devices)
    throw (Opencl_error)
{
	cl_int	err;
	bool	verbose = false;

	// get platforms
	std::vector<cl::Platform> platforms;
	err = cl::Platform::get(&platforms);
	opencl_check("getting platforms", err);

	// get devices
	std::vector<cl::Device> devices;
	for (size_t i = 0; i < platforms.size(); i++) {
		std::string platform;

		if (verbose) {
			err = platforms[i].getInfo(CL_PLATFORM_NAME,
			    &platform);
			opencl_check("getting platform name", err);
		}

		std::vector<cl::Device> pf_devices;
		err = platforms[i].getDevices(CL_DEVICE_TYPE_GPU, &pf_devices);
		opencl_check("getting devices", err);

		for (size_t j = 0; verbose && j < pf_devices.size(); j++) {
			std::string device;
			err = pf_devices[j].getInfo(CL_DEVICE_NAME, &device);
			opencl_check("getting device name", err);
			std::cout << platform << " - " << device << '\n';
		}

		// concatenate onto devices
		std::copy(pf_devices.begin(), pf_devices.end(),
		    std::back_inserter(devices));
	}

	// create context
	cl::Context ctx(devices, 0, 0, 0, &err);
	opencl_check("creating context", err);

	std::copy(devices.begin(), devices.end(),
	    std::back_inserter(out_devices));
	return ctx;
}

void
create_opencl_kernels(cl::Context &ctx,
    const std::vector<cl::Device> &devices,
    const std::vector<std::string> &source_paths,
    std::map<std::string, cl::Kernel> &kernels) throw (Opencl_error)
{
	cl_int	err;

	// read source code
	std::vector<std::string> source_code(source_paths.size());
	std::vector<std::pair<const char *, size_t> > sources;
	for (size_t i = 0; i < source_paths.size(); i++) {
		read_file(source_paths[i], source_code[i]);
		sources.push_back(std::make_pair(source_code[i].c_str(), 0));
	}

	// create program
	cl::Program program(ctx, sources, &err);
	opencl_check("creating program", err);

	// build code
	err = program.build(devices, "-I.", 0, 0);
	if (err == CL_BUILD_PROGRAM_FAILURE) {
		std::string build_log;
		err = program.getBuildInfo(devices[0], CL_PROGRAM_BUILD_LOG,
		    &build_log);
		opencl_check("getting build log", err);
		if (!build_log.empty())
			std::cout << "build log:\n" << build_log;
		opencl_check("building program", CL_BUILD_PROGRAM_FAILURE);
	} else
		opencl_check("building program", err);

	// display build log if length>0
	std::string build_log;
	err = program.getBuildInfo(devices[0], CL_PROGRAM_BUILD_LOG,
	    &build_log);
	opencl_check("getting build log", err);
	if (!build_log.empty())
		std::cout << "build log:\n" << build_log;

	// just a hint
	//err = clUnloadCompiler();
	opencl_check("unloading compiler", err);

	// create kernels
	std::vector<cl::Kernel> kernel_vec;
	err = program.createKernels(&kernel_vec);
	opencl_check("creating kernels", err);

	// output to the map argument
	for (size_t i = 0; i < kernel_vec.size(); i++) {
		std::string name;
		err = kernel_vec[i].getInfo(CL_KERNEL_FUNCTION_NAME, &name);
		opencl_check("getting kernel name", err);
		kernels[name] = kernel_vec[i];
	}
}

void
read_file(const std::string &pathname, std::string &out)
{
	std::ifstream in(pathname.c_str());
	if (!in) {
		// FIXME this could be better
		std::cerr << "file doesn't exist\n";
		assert(0);
	}

	char buf[4096];
	while (in) {
		in.read(buf, sizeof(buf));
		if (in.bad()) break;
		size_t bytes = in.gcount();
		out.append(buf, bytes);
	}
}

} // end anon namespace

void
Openclk::init(const std::vector<std::string> &src_paths)
    throw (Opencl_error)
{
	cl_int err;

	std::vector<cl::Device> devices;
	_ctx = create_opencl_context(devices);
	create_opencl_kernels(_ctx, devices, src_paths, _kernels);

	_queue = cl::CommandQueue(_ctx, devices[0],
	    CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE, &err);
	opencl_check("creating queue", err);

	for (size_t i = 0; i < devices.size(); i++) {
		cl_command_queue_properties props;
		props = devices[i].getInfo<CL_DEVICE_QUEUE_PROPERTIES>(&err);
		opencl_check("getting device queue properties", err);
		if (!(props & CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE)) {
			std::cerr
			    << "warning: queue reordering not supported\n";
			break;
		}
	}

	_max_group = devices[0].getInfo<CL_DEVICE_MAX_WORK_GROUP_SIZE>(&err);
	opencl_check("getting device properties", err);

	err = devices[0].getInfo(CL_DEVICE_MAX_WORK_ITEM_SIZES,
	    &_max_work_sizes);
	opencl_check("getting device properties", err);
}
