/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef OPENCLK_HPP
#define OPENCLK_HPP

#include <cstdlib>
#include <map>
#include <string>
#include <vector>
#include <tr1/memory>

#include <CL/cl.hpp>

#include "errors.hpp"
#include "log.hpp"

// to identify host/device in a shared header
#define OPENCLK_HOST

inline cl::Event	*convert_basic(cl_event &);
inline cl::Kernel	*convert_basic(cl_kernel &);
// define the rest similarly...

class Task_group;

class Openclk {
public:
	Openclk(const std::vector<std::string> &src_paths)
	    throw (Opencl_error)
	{
		init(src_paths);
	}

	template <typename In>
	Openclk(In src_path_begin, In src_path_end)
	    throw (Opencl_error)
	{
		std::vector<std::string> src_paths(src_path_begin,
		    src_path_end);
		init(src_paths);
	}

	cl::Context &context()
	{
		return _ctx;
	}
	cl::Kernel &kernel(const std::string &name)
	{
		return _kernels[name];
	}
	cl::CommandQueue &queue()
	{
		return _queue;
	}

	const std::vector<size_t> &max_work_sizes() const
	{	return _max_work_sizes; }

	size_t max_group() const
	{	return _max_group; }

private:
	void init(const std::vector<std::string> &src_paths)
	    throw (Opencl_error);

	// non-copyable
	Openclk(const Openclk &) { throw std::exception(); }
	void operator=(const Openclk &) { throw std::exception(); }

	std::map<std::string, cl::Kernel>	_kernels;
	cl::Context		_ctx;
	cl::CommandQueue	_queue;
	std::vector<size_t>	_max_work_sizes;
	size_t			_max_group;
};

class Argument {
public:
	virtual ~Argument() {}

	virtual std::pair<void *, size_t> get() = 0;
	virtual bool is_buffer() const = 0;

	// only relevant for Argument_buffer:
	virtual bool readable() const
	{
		return true;
	}
	virtual bool writable() const
	{
		return false;
	}
};

template <typename T>
class Argument_value : public Argument {
public:
	Argument_value(T val) :
		_val(val)
	{}

	virtual std::pair<void *, size_t> get()
	{
		return std::pair<void *, size_t>(&_val, sizeof(T));
	}
	virtual bool is_buffer() const
	{
		return false;
	}

private:
	T _val;
};

class Argument_buffer : public Argument {
public:
	enum access { read=0, write=1, read_write=2 };

	Argument_buffer(void *buf, size_t sz, enum access access) :
		_buf(buf),
		_sz(sz),
		_access(access)
	{}

	virtual std::pair<void *, size_t> get()
	{
		return std::pair<void *, size_t>(_buf, _sz);
	}
	virtual bool is_buffer() const
	{
		return true;
	}
	virtual bool readable() const
	{
		return _access == read || _access == read_write;
	}
	virtual bool writable() const
	{
		return _access == write || _access == read_write;
	}

private:
	void *_buf;
	size_t _sz;
	enum access _access;
};

template <typename Result>
class Alg_spec {
public:
	typedef Result result_type;
	typedef std::vector<std::tr1::shared_ptr<Argument> > arg_vector;

	Alg_spec() {}

	virtual ~Alg_spec() {
	    printLog();
	}

	/**
	 * Add the initial value for an argument, defined only for OpenCL
	 * primitives.
	 */
	void add_arg(cl_char x)
	{	_add_val(x, _args); }
	void add_arg(cl_uchar x)
	{	_add_val(x, _args); }

	void add_arg(cl_short x)
	{	_add_val(x, _args); }
	void add_arg(cl_ushort x)
	{	_add_val(x, _args); }

	void add_arg(cl_int x)
	{	_add_val(x, _args); }
	void add_arg(cl_uint x)
	{	_add_val(x, _args); }

	void add_arg(cl_long x)
	{	_add_val(x, _args); }
	void add_arg(cl_ulong x)
	{	_add_val(x, _args); }

	void add_arg(cl_float x)
	{	_add_val(x, _args); }
	void add_arg(cl_double x)
	{	_add_val(x, _args); }

	/**
	 * Add the initial value for an buffer argument, defined only for
	 * OpenCL primitives.
	 */
	void add_arg(const cl_char *buf, size_t sz)
	{
		_add_buf(const_cast<cl_char *>(buf), sz,
		    Argument_buffer::read, _args);
	}
	void add_arg(const cl_uchar *buf, size_t sz)
	{
		_add_buf(const_cast<cl_uchar *>(buf), sz,
		    Argument_buffer::read, _args);
	}

	void add_arg(const cl_short *buf, size_t sz)
	{
		_add_buf(const_cast<cl_short *>(buf), sz,
		    Argument_buffer::read, _args);
	}
	void add_arg(const cl_ushort *buf, size_t sz)
	{
		_add_buf(const_cast<cl_ushort *>(buf), sz,
		    Argument_buffer::read, _args);
	}

	void add_arg(const cl_int *buf, size_t sz)
	{
		_add_buf(const_cast<cl_int *>(buf), sz,
		    Argument_buffer::read, _args);
	}
	void add_arg(const cl_uint *buf, size_t sz)
	{
		_add_buf(const_cast<cl_uint *>(buf), sz,
		    Argument_buffer::read, _args);
	}

	void add_arg(const cl_long *buf, size_t sz)
	{
		_add_buf(const_cast<cl_long *>(buf), sz,
		    Argument_buffer::read, _args);
	}
	void add_arg(const cl_ulong *buf, size_t sz)
	{
		_add_buf(const_cast<cl_ulong *>(buf), sz,
		    Argument_buffer::read, _args);
	}

	void add_arg(const cl_float *buf, size_t sz)
	{
		_add_buf(const_cast<cl_float *>(buf), sz,
		    Argument_buffer::read, _args);
	}
	void add_arg(const cl_double *buf, size_t sz)
	{
		_add_buf(const_cast<cl_double *>(buf), sz,
		    Argument_buffer::read, _args);
	}

	/**
	 * Add the initial value for a generic argument, which *should* be
	 * composed of OpenCL primitives only.
	 */
	void add_arg_generic(const void *data, size_t sz)
	{
		_add_buf(const_cast<void *>(data), sz, Argument_buffer::read,
		    _args);
	}

	/**
	 * Add the value of a primitive parameter that will not change between
	 * tasks, defined only for OpenCL primitives.
	 */
	void add_param(cl_char x)
	{	_add_val(x, _params); }
	void add_param(cl_uchar x)
	{	_add_val(x, _params); }

	void add_param(cl_short x)
	{	_add_val(x, _params); }
	void add_param(cl_ushort x)
	{	_add_val(x, _params); }

	void add_param(cl_int x)
	{	_add_val(x, _params); }
	void add_param(cl_uint x)
	{	_add_val(x, _params); }

	void add_param(cl_long x)
	{	_add_val(x, _params); }
	void add_param(cl_ulong x)
	{	_add_val(x, _params); }

	void add_param(cl_float x)
	{	_add_val(x, _params); }
	void add_param(cl_double x)
	{	_add_val(x, _params); }

	/**
	 * Add the value of a buffer parameter that will not change between
	 * tasks, defined only for OpenCL primitives.
	 */
	void add_param(cl_char *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }
	void add_param(cl_uchar *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }

	void add_param(cl_short *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }
	void add_param(cl_ushort *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }

	void add_param(cl_int *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }
	void add_param(cl_uint *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }

	void add_param(cl_long *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }
	void add_param(cl_ulong *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }

	void add_param(cl_float *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }
	void add_param(cl_double *buf, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(buf, sz, access, _params); }

	/**
	 * Add the initial value for a generic parameter, which *should* be
	 * composed of OpenCL primitives only.
	 */
	void add_param_generic(void *data, size_t sz,
	    enum Argument_buffer::access access)
	{	_add_buf(data, sz, access, _params); }

	template <typename T>
	void add_param_generic(const T &t)
	{	_add_val(t, _params); }

	/**
	 * Run the task.  This will return the result of the root task as well
	 * as write out the final values of any writable parameters.
	 */
	Result run(Openclk &openclk) throw (Opencl_error);

protected:
	/**
	 * The name of the kernel function.
	 */
	virtual std::string name() const = 0;

	/**
	 * The reduction function, which is executed on the CPU due to its
	 * simplicity.
	 *
	 * \param child_results Results from spawned tasks
	 * \param len The number of spawned tasks; it should always be
	 *	sub_tasks() come to think of it
	 * \param result The result of the reduction; the return value of the
	 *	task
	 */
	virtual void reduce(const result_type *child_results, size_t len,
	    result_type *result) = 0;

	/**
	 * The number of subtasks.  In Cilk, this would be the number of
	 * times spawn() would appear.
	 */
	virtual unsigned sub_tasks() const = 0;

	size_t result_size() const
	{	return sizeof(result_type); }

private:
	void reduce_group(Task_group &group, unsigned task,
	    result_type *responses, unsigned sub_tasks);

	template <typename T>
	void _add_val(const T &x, arg_vector &vec)
	{
		vec.push_back(std::tr1::shared_ptr<Argument>(
		    new Argument_value<T>(x)));
	}

	void _add_buf(void *buf, size_t sz,
	    enum Argument_buffer::access access, arg_vector &vec)
	{
		vec.push_back(std::tr1::shared_ptr<Argument>(
		    new Argument_buffer(buf, sz, access)));
	}

	arg_vector	_args;
	arg_vector	_params;
};

#include "openclk_private.hpp"

#endif
