/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <limits>
#include <set>
#include <sstream>

#include "task.hpp"

inline cl::Event *
convert_basic(cl_event &e)
{
	// the only member of cl::Event is a cl_event; this is the hack that
	// the cl.hpp authors used
	return reinterpret_cast<cl::Event *>(&e);
}

inline cl::Kernel *
convert_basic(cl_kernel &k)
{
	return reinterpret_cast<cl::Kernel *>(&k);
}

template <typename Result>
Result
Alg_spec<Result>::run(Openclk &openclk) throw (Opencl_error)
{
	typedef std::map<cl_event, std::tr1::shared_ptr<Task_group> >
	    group_map;
	typedef std::map<Task_group *, cl_event> inverted_map;

	// pairs of Buffer, index (into _params)
	std::vector<std::pair<cl::Buffer, size_t> > write_buffers;
	// keep these from being destroyed
	std::vector<cl::Buffer> read_buffers;
	Queue<cl_event>		comp_queue;
	std::set<Task_group *>	running;
	std::deque<std::pair<Task_group *, size_t> > waiting;
	group_map		groups;
	inverted_map		inverse;

	cl::Context		&ctx = openclk.context();
	cl::Kernel		&kernel = openclk.kernel(name());
	cl::CommandQueue	&queue = openclk.queue();
	cl_int			err;
	// save result of virtual call
	unsigned		sub_tasks = this->sub_tasks();

	// the most tasks that can be executed at once
	unsigned	max_tasks = std::numeric_limits<unsigned>::max();

	// create the root task
	logEvent("start memory transfer");
	std::tr1::shared_ptr<Root_task> root(new Root_task(
	    openclk.context(), sub_tasks, result_size()));

	// set the parameters (that don't change ever)
	for (size_t i = 0; i < _params.size(); i++) {
		std::pair<void *, size_t> param = _params[i]->get();
		if (_params[i]->is_buffer()) {
			cl_int access;
			if (_params[i]->readable())
				if (_params[i]->writable())
					access = CL_MEM_READ_WRITE |
					    CL_MEM_COPY_HOST_PTR;
				else
					access = CL_MEM_READ_ONLY |
					    CL_MEM_COPY_HOST_PTR;
			else {
				param.first = 0;
				access = CL_MEM_WRITE_ONLY |
				    CL_MEM_ALLOC_HOST_PTR;
			}

			cl::Buffer buf(ctx, access, param.second, param.first,
			    &err);
			if (err == CL_INVALID_HOST_PTR) {
				std::ostringstream out;
				out << "creating parameter buffer, index="
				    << i << ", host_ptr"
				    << (param.first ? "!=" : "==")
				    << "0, flags=" << access;
				opencl_check(out.str(), err);
			} else
				opencl_check("creating parameter buffer",
				    err);

			err = kernel.setArg(i, buf);
			opencl_check("setting kernel buffer parameter", err);

			if (_params[i]->writable())
				write_buffers.push_back(std::make_pair(buf,
				    i));
			else
				read_buffers.push_back(buf);
		} else {
			// C-style since Argument::get() can't include type
			// information
			//std::cerr << "set arg " << i << " to value at " << param.first << '\n';
			err = clSetKernelArg(kernel(), i,
			    param.second, param.first);
			if (err) {
				std::cerr << "size: " << param.second << '\n';
			}
			opencl_check("setting kernel value parameter", err);
		}
	}
	logEvent("finish memory transfer");

	// arguments are task-group dependent
	for (size_t i = 0; i < _args.size(); i++) {
		std::pair<const void *, size_t> arg = _args[i]->get();
		assert(!(_args[i]->is_buffer() && _args[i]->writable()));
		root->add_arg_generic(arg.first, arg.second);
	}

	// if there were constant kernel arguments: kernel.setArg() ...

	cl::Event complete = root->invoke(queue, kernel, comp_queue,
	    openclk.max_group(), _params.size());
	groups[complete()] = root;
	inverse[root.get()] = complete();
	running.insert(root.get());

	result_type *reduce_buf = new result_type[sub_tasks];

	bool have_resp = false;
	result_type final_resp = 0;
	for (;;) {
		cl_event evt_basic = comp_queue.pop();
		//cl::Event *evt = convert_basic(evt_basic);

		std::tr1::shared_ptr<Task_group> group = groups[evt_basic];
		running.erase(group.get());

		group->reorder_buffers(sub_tasks, root->arg_sizes());

		while (!waiting.empty() && running.size() != max_tasks) {
			// room in the run queue for one more

			std::pair<Task_group *, size_t> parent =
			    waiting.front();
			waiting.pop_front();

			// attempt to create subtasks of 'parent'
			logEvent("start memory transfer");
			std::tr1::shared_ptr<Child_group> child(
			    new Child_group(openclk.context(), result_size(),
			    *root, *parent.first, parent.second));
			logEvent("finish memory transfer");
			complete = child->invoke(queue, kernel, comp_queue,
			    openclk.max_group(), _params.size());
			if (complete()) {
				// task was queued
				groups[complete()] = child;
				inverse[child.get()] = complete();
				running.insert(child.get());
			} else {
				max_tasks = running.size();
				std::cerr << "Command queue at its limit: "
				    << max_tasks << " task groups.  "
				    "You should give up now.\n";
				// re-add to wait queue instead
				waiting.push_front(parent);
			}
		}

		group->release_gpu_buffers();

#ifdef QUEUE_DEBUG
		std::ostringstream out;
		out << "queue :: retrieved " << evt_basic << '\n'
		    << "      :: group " << group.get() << '\n'
		    << "      :: depth " << group->depth() << '\n';
		std::cerr << out.str();
#endif

		// for each subtask of parent...
		size_t tasks = group->tasks();
		/*
		for (i = 0; i < tasks; i++) {
			if (!group->spawn(i))
				// done with spawning tasks
				break;

			if (running.size() == max_tasks) {
				// just add to wait queue
				waiting.push_back(std::make_pair(
				    group.get(), i));
				continue;
			}

			// ...create further subtasks
			std::tr1::shared_ptr<Child_group> child(
			    new Child_group(openclk.context(), result_size(),
				*root, *group, i));

			complete = child->invoke(queue, kernel, comp_queue,
			    openclk.max_group(), _params.size());
			if (complete()) {
				// task was queued
				groups[complete()] = child;
				inverse[child.get()] = complete();
				running.insert(child.get());
			} else {
				max_tasks = running.size();
				std::cerr << "Command queue at its limit: "
				    << max_tasks << " task groups.  "
				    "You should give up now.\n";
				// add to wait queue instead; back/front
				// doesn't matter, it should be empty
				waiting.push_back(std::make_pair(
				    group.get(), i));
			}
		}
		*/

		// find start of non-spawned tasks
		size_t i;
		for (i = 0; i < tasks; i++)
			if (!group->spawn(i))
				break;

		if (i) {
			// ...create further subtasks
			logEvent("start memory transfer");
			std::tr1::shared_ptr<Child_group> child(
			    new Child_group(openclk.context(), result_size(),
				*root, *group, 0));
			logEvent("finish memory transfer");

			complete = child->invoke(queue, kernel, comp_queue,
			    openclk.max_group(), _params.size());
			if (complete()) {
				// task was queued
				groups[complete()] = child;
				inverse[child.get()] = complete();
				running.insert(child.get());
			} else {
				assert(0);
			}
		}

		logEvent("start reducing tasks");
		// process non-spawning tasks
		//std::cerr << "reduce loop\n";
		for (; i < tasks; i++) {
			Task_group *current = group.get();
			unsigned offset = i;
			do {
				Task_group	*parent = current->parent();
				unsigned	parent_offset =
				    current->parent_offsets()[offset];

				//std::cerr << "reduce (" << current << ',' << offset << ") parent (" << parent << ',' << parent_offset << ")\n";

				if (parent) {
					unsigned completed =
					    parent->feed_result(parent_offset,
					    current->response(offset));
					if (completed != sub_tasks)
						break;
					// else: can reduce
					//std::cerr << "REDUCE\n";

					reduce_group(*current, offset,
					    reduce_buf, sub_tasks);
				} else {
					result_type	*current_resp =
					    reinterpret_cast<result_type *>(
					    current->response());
					// signal completion
					final_resp = *current_resp;
					have_resp = true;
					break;
				}

				// destroy the group
				/* FIXME
				cl_event current_event = inverse[current];
				inverse.erase(current);
				groups.erase(current_event);
				// don't need the event anymore
				clReleaseEvent(current_event);
				*/

				current = parent;
				offset = parent_offset;
			} while (1);//!current->all_resolved());

			if (have_resp)
				// complete
				break;
		}
		logEvent("finish reducing tasks");
		if (have_resp)
			// complete
			break;
	}
	delete[] reduce_buf;

	// copy writable buffers back
	std::vector<cl::Event> events;
	for (size_t i = 0; i < write_buffers.size(); i++) {
		size_t j = write_buffers[i].second;
		std::pair<void *, size_t> dest = _params[j]->get();
		bool last = i == write_buffers.size() - 1;

		// enqueue the read, block the last time,
		cl::Event complete;
		err = queue.enqueueReadBuffer(write_buffers[i].first, last,
		    0, dest.second, dest.first, &events, &complete);
		opencl_check("enquing parameter read", err);

		if (!last) {
			events.resize(1);
			events[0] = complete;
		}
	}

	return final_resp;
}

template <typename Result>
void
Alg_spec<Result>::reduce_group(Task_group &group, unsigned task,
    result_type *responses, unsigned sub_tasks)
{
	const std::vector<unsigned> &parent_offsets = group.parent_offsets();
	Task_group	*parent = group.parent();

	// copy the results to a single buffer
	std::vector<const void *> &child_results =
	    parent->child_results(parent_offsets[task]);
	result_type *dst = responses;
	for (size_t i = 0; i < sub_tasks; i++) {
		const result_type *src = reinterpret_cast<
		    const result_type *>(child_results[i]);
		*dst++ = *src;
	}
	child_results = std::vector<const void *>();

	// parent_resp: where to write reduction result
	result_type *parent_resp =
	    reinterpret_cast<result_type *>(
	    parent->response(parent_offsets[task]));
	reduce(responses, sub_tasks, parent_resp);
	parent->mark_resolved(parent_offsets[task]);
}
