/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#include <iostream>
#include <sstream>

#include "task.hpp"
#include "log.hpp"

namespace {

void		event_complete(cl_event, cl_int, void *);


void
event_complete(cl_event event, cl_int command_exec_status,
    void *completion_queue)
{
	Queue<cl_event> *comp_queue =
	    reinterpret_cast<Queue<cl_event> *>(completion_queue);

	logGPUEvent("finish kernel invocation");
	comp_queue->push(event);

#ifdef QUEUE_DEBUG
	std::ostringstream out;
	out << "queue :: completed " << event << '\n';
	std::cerr << out.str();
#endif
}

} // end anon namespace

Task_group::Task_group(cl::Context &ctx, unsigned tasks, unsigned sub_tasks,
    size_t response_sz, Task_group *parent, unsigned parent_offset)
    throw (Opencl_error) :
	_args(),
	_spawn_args(),
	_spawn_arg_bufs(),
	_parent_offsets(tasks),
	_child_results(tasks),
	_response(),
	_spawn(),
	_response_buf(0),
	_spawn_buf(0),
	_response_sz(response_sz),
	_parent(parent),
	_num_spawns(-1)
{
	cl_int	err;
	size_t	all_responses = response_sz * tasks;
	assert(tasks);

	try {

		// length = ceil(all_responses / 8)
		_response_buf = new cl_ulong[(all_responses + 7) / 8];
		_spawn_buf = new cl_uint[tasks];

		_response = cl::Buffer(ctx,
		    CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR,
		    all_responses, 0, &err);
		opencl_check("creating response buffer", err);

		_spawn = cl::Buffer(ctx,
		    CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR,
		    sizeof(cl_uint) * tasks, 0, &err);
		opencl_check("creating spawn buffer", err);
	} catch (...) {
		if (_response_buf) delete[] _response_buf;
		if (_spawn_buf) delete[] _spawn_buf;
		throw;
	}

	unsigned group = 0;
	unsigned left = sub_tasks;
	for (unsigned i = 0; i < tasks; i++) {
		_parent_offsets[i] = group + parent_offset;
		if (!--left) {
			left = sub_tasks;
			group++;
		}
	}

#ifdef QUEUE_DEBUG
	_depth = parent ? parent->_depth + 1 : 0;
#endif
}

cl::Event
Task_group::invoke(cl::CommandQueue &queue, cl::Kernel &kernel,
    Queue<cl_event> &complete, cl_uint group_size, unsigned first_arg_idx)
    throw (Opencl_error)
{
	std::vector<cl::Event> events(1);
	cl::Event kernel_complete;
	cl::Event result_complete;
	cl::Event spawn_complete;
	cl_int err;

	cl_uint tasks = this->tasks();
	cl_uint global = tasks;
	if (global < group_size)
		group_size = global;
	else
		global = (global + group_size - 1) / group_size * group_size;

	// max_tid
	kernel.setArg(first_arg_idx++, tasks-1);

	// set the variable kernel arguments
	for (size_t i = 0; i < _args.size(); i++) {
		err = kernel.setArg(first_arg_idx++, _args[i]);
		opencl_check("setting kernel arg", err);
	}
	err = kernel.setArg(first_arg_idx++, _response);
	opencl_check("setting kernel arg", err);
	err = kernel.setArg(first_arg_idx++, _spawn);
	opencl_check("setting kernel arg", err);
	for (size_t i = 0; i < _spawn_args.size(); i++) {
		err = kernel.setArg(first_arg_idx++, _spawn_args[i]);
		opencl_check("setting kernel arg", err);
	}

	//std::cerr << tasks << ' ' << group_size << '\n';

	// invoke kernel
	logEvent("start kernel invocation");
	err = queue.enqueueNDRangeKernel(kernel,
	    cl::NDRange(0), // offset
	    cl::NDRange(global), // global size
	    cl::NDRange(group_size), // local size
	    0, &kernel_complete);
	if (err == CL_OUT_OF_RESOURCES)
		return cl::Event();
	opencl_check("enquing kernel", err);

	// read result
	events[0] = kernel_complete;
	err = queue.enqueueReadBuffer(_response, false, 0,
	    _response_sz * tasks, _response_buf, &events, &result_complete);
	opencl_check("enquing read", err);

	// read spawn
	events[0] = result_complete;
	err = queue.enqueueReadBuffer(_spawn, false, 0,
	    sizeof(cl_uint) * tasks, _spawn_buf, &events, &spawn_complete);
	opencl_check("enquing read", err);

	// enqueue reads of all spawn arguments
	cl::Event last = spawn_complete;
	for (size_t i = 0; i < _spawn_args.size(); i++) {
		cl::Event next;
		events[0] = last;
		err = queue.enqueueReadBuffer(_spawn_args[i], false, 0,
		    _spawn_arg_bufs[i].second, _spawn_arg_bufs[i].first,
		    &events, &next);
		last = next;
	}

	// set callback
	err = last.setCallback(CL_COMPLETE, event_complete, &complete);
	opencl_check("setting completion callback", err);

	// we don't want the event to be reused since it's the index to a
	// map
	clRetainEvent(last());

	// start
	queue.flush();
	logEvent("start asynchronous call");
	return last;
}

void
Task_group::reorder_buffers(unsigned sub_tasks,
    const std::vector<size_t> &arg_sizes)
{
	logEvent("start reordering buffers");
	size_t i, j;
	uint8_t *dst, *src;
	unsigned tasks = this->tasks();

	/*cl_uint *resp = reinterpret_cast<cl_uint *>(_response_buf);
	std::cout << "before\n";
	for (i = 0; i < tasks; i++)
		std::cout << _spawn_buf[i] << ' ';
	std::cout << '\n';
	for (i = 0; i < tasks; i++)
		std::cout << resp[i] << ' ';
	std::cout << '\n';*/

	// rearrange response, parent_offsets; this will bubble responses
	// towards the right
	i = j = 0;
	dst = src = reinterpret_cast<uint8_t *>(_response_buf);
	while (j < tasks) {
		if (_spawn_buf[j]) {
			if (i != j) {
				// responses have to move left->right, so
				// dst->src
				std::copy(dst, dst + _response_sz, src);
				std::swap(_parent_offsets[i],
				    _parent_offsets[j]);
			}
			dst += _response_sz;
			i++;
		}
		src += _response_sz;
		j++;
	}

	// rearrange each spawn_args
	for (size_t k = 0; k < _spawn_arg_bufs.size(); k++) {
		i = j = 0;
		dst = src = reinterpret_cast<uint8_t *>(
		    _spawn_arg_bufs[k].first);
		size_t portion = arg_sizes[k] * sub_tasks;
		while (j < tasks) {
			if (_spawn_buf[j]) {
				if (i != j)
					std::copy(src, src + portion, dst);
				dst += portion;
				i++;
			}
			src += portion;
			j++;
		}
	}

	// rearrange _spawn_buf
	i = j = 0;
	while (j < tasks) {
		if (_spawn_buf[j]) {
			if (i != j) {
				_spawn_buf[i] = true;
				_spawn_buf[j] = false;
			}
			i++;
		}
		j++;
	}

	/*std::cout << "after\n";
	for (i = 0; i < tasks; i++)
		std::cout << _spawn_buf[i] << ' ';
	std::cout << '\n';
	for (i = 0; i < tasks; i++)
		std::cout << resp[i] << ' ';
	std::cout << '\n';*/

	logEvent("finish reordering buffers");
}

unsigned
Task_group::num_spawn_groups() const
{
	if (_num_spawns == -1) {
		Task_group *this_ = const_cast<Task_group *>(this);
		size_t tasks = this->tasks();
		this_->_num_spawns = 0;
		for (size_t i = 0; i < tasks; i++)
			if (_spawn_buf[i]) this_->_num_spawns++;
	}
	return _num_spawns;
}

// create two GPU buffers: one for input to the task, one for input to
// subtasks; the latter gets a corresponding CPU buffer; record the size of
// the argument in _arg_sizes
void
Root_task::add_arg_generic(const void *val, size_t size) throw (Opencl_error)
{
	cl_int err;

	// create input argument
	_args.push_back(cl::Buffer(_ctx,
	    CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, size,
	    const_cast<void *>(val), &err));
	opencl_check("adding argument", err);

	size_t sz_spawn_arg = size * _num_tasks;

	// create GPU buffer for output
	_spawn_args.push_back(cl::Buffer(_ctx,
	    CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR, sz_spawn_arg, 0, &err));
	opencl_check("adding spawn buffer", err);

	// create CPU buffer for output
	_spawn_arg_bufs.push_back(std::make_pair(
	    new cl_ulong[(sz_spawn_arg + 7) / 8], sz_spawn_arg));

	// record size
	_arg_sizes.push_back(size);
}

Child_group::Child_group(cl::Context &ctx, size_t response_sz,
    const Root_task &root, Task_group &parent_group, unsigned parent_offset)
    throw (Opencl_error) :
	Task_group(ctx, parent_group.num_spawn_groups() * root.child_tasks(),
	    root.child_tasks(), response_sz, &parent_group, parent_offset),
	_num_tasks(parent_group.num_spawn_groups() * root.child_tasks()),
	_num_child_tasks(_num_tasks * root.child_tasks())
{
	cl_int err;
	const std::vector<size_t> &arg_sizes = root.arg_sizes();

	_args.resize(arg_sizes.size());
	_spawn_args.resize(arg_sizes.size());
	_spawn_arg_bufs.resize(arg_sizes.size());

	// for each argument
	for (unsigned i = 0; i < arg_sizes.size(); i++) {
		size_t arg_size = arg_sizes[i];

		// the size of an input arg buffer (for tasks)
		size_t sz_buffer = arg_size * _num_tasks;
		// the size of an output arg buffer (for subtasks)
		size_t sz_nbuffers = sz_buffer * _num_child_tasks;

		// the parent's spawn_arg_buf for this argument
		uint8_t *parent_buf = reinterpret_cast<uint8_t *>(
		    parent_group.spawn_arg_buf(i));

#if 0 // fib only
		cl_uint *b = reinterpret_cast<cl_uint *>(parent_buf +
		    sz_buffer * parent_offset);
		std::cerr << "will call fib(" << b[0] << "), fib(" << b[1]
		    << ")\n";
#endif

#if 0 // heat only
		cl_uint *b = reinterpret_cast<cl_uint *>(parent_buf +
		    sz_buffer * parent_offset);
		std::cerr << "will call heat with arg " << i
		    << " = {" << b[0] << ',' << b[1] << "}\n";

		b = reinterpret_cast<cl_uint *>(parent_group.response());
		std::cerr << "leaf_max_col was " << b[parent_offset] << '\n';
#endif

#if 0 // matmul only
		cl_uint *b = reinterpret_cast<cl_uint *>(parent_buf +
		    sz_buffer * parent_offset);
		std::cerr << "will call matmul with arg " << i
		    << " = {" << b[0] << ',' << b[1] << ", ...}\n";

		b = reinterpret_cast<cl_uint *>(parent_group.response());
		std::cerr << "diff " << b[parent_offset] << '\n';
#endif

		_args[i] = cl::Buffer(ctx,
		    CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
		    sz_buffer, parent_buf + sz_buffer * parent_offset,
		    &err);
		opencl_check("creating args buffer", err);

		/* XXX what subregion construction looks like:

		cl_buffer_region sub_region;
		sub_region.origin = sz_buffer * parent_offset;
		sub_region.size = sz_buffer;

		//std::cerr << "sub region [" << sub_region.origin << ','
		//    << sub_region.origin + sub_region.size << ")\n";

		// args is a subregion of the parent's spawn_args
		_args[i] = parent_group.spawn_arg(i).createSubBuffer(
		    CL_MEM_READ_ONLY,
		    CL_BUFFER_CREATE_TYPE_REGION, &sub_region,
		    &err);
		opencl_check("creating args sub-buffer", err);
		*/

		// allocate memory for spawn_args
		_spawn_args[i] = cl::Buffer(ctx,
		    CL_MEM_WRITE_ONLY | CL_MEM_ALLOC_HOST_PTR,
		    sz_nbuffers, 0, &err);
		if (err) {
			std::cerr << "buffer size: " << sz_nbuffers << '\n';
		}
		opencl_check("creating spawn_args buffer", err);

		_spawn_arg_bufs[i] = std::make_pair(
		    new cl_ulong[(sz_nbuffers + 7) / 8], sz_nbuffers);
	}
}
