/* Copyright 2010 Markus Peloquin <markus@cs.wisc.edu>,
 * Lena Olson <lena@cs.wisc.edu>
 *
 * You can't do anything with this but read and execute it privately.  If you
 * want to do more, ask us if we'd be willing to make a more permissive
 * license. */

#ifndef OPENCLK_TASK_HPP
#define OPENCLK_TASK_HPP

#include <cassert>

#include <CL/cl.hpp>

#include "errors.hpp"
#include "queue.hpp"

class Task_group {
protected:
	Task_group(cl::Context &ctx, unsigned tasks, unsigned sub_tasks,
	    size_t response_sz, Task_group *parent, unsigned parent_offset)
	    throw (Opencl_error);

public:
	virtual ~Task_group()
	{
		for (size_t i = 0; i < _spawn_arg_bufs.size(); i++)
			delete[] _spawn_arg_bufs[i].first;
		delete[] _response_buf;
		delete[] _spawn_buf;
	}

	cl::Event invoke(cl::CommandQueue &queue, cl::Kernel &kernel,
	    Queue<cl_event> &complete, cl_uint group_size,
	    unsigned first_arg_idx) throw (Opencl_error);

	// when the result of invoke() is pushed to the completion queue, this
	// should be called to release the unneeded arguments (not
	// spawn_args since that's still being used)
	void release_gpu_buffers()
	{
		for (size_t i = 0; i < _args.size(); i++) {
			_args[i] = cl::Buffer();
			_spawn_args[i] = cl::Buffer();
		}
		_response = cl::Buffer();
		_spawn = cl::Buffer();
	}

	cl::Buffer &spawn_arg(unsigned arg)
	{
		assert(arg < _spawn_args.size());
		return _spawn_args[arg];
	}
	void *spawn_arg_buf(unsigned arg)
	{
		assert(arg < _spawn_arg_bufs.size());
		return _spawn_arg_bufs[arg].first;
	}
	void *response()
	{
		return _response_buf;
	}
	void *response(unsigned task)
	{
		assert(task < tasks());
		return reinterpret_cast<uint8_t *>(_response_buf) +
		    task * _response_sz;
	}

	bool spawn(unsigned task) const
	{
		assert(task < tasks());
		assert(_spawn_buf[task] < 2);
		return _spawn_buf[task];
	}
	void mark_resolved(unsigned task)
	{
		assert(task < tasks());
		_spawn_buf[task] = 0;
	}
	bool all_resolved() const
	{
		size_t tasks = this->tasks();
		for (size_t i = 0; i < tasks; i++)
			if (_spawn_buf[i])
				return true;
		return false;
	}

	virtual unsigned tasks() const = 0;
	// the maximum number of spawned tasks
	virtual unsigned child_tasks() const = 0;
	// valid only after the task group completes
	unsigned num_spawn_groups() const;

	Task_group *parent()
	{	return _parent; }
	const std::vector<unsigned> &parent_offsets() const
	{	return _parent_offsets; }

	// reorders _response_buf, _spawn_buf, _spawn_arg_bufs[i],
	// _parent_offsets so that _spawn_arg_bufs[i] will be contiguous
	void reorder_buffers(unsigned sub_tasks,
	    const std::vector<size_t> &arg_sizes);

	// returns number of results so far obtained for this task
	unsigned feed_result(unsigned task, const void *result)
	{
		assert(task < _child_results.size());
		_child_results[task].push_back(result);
		return _child_results[task].size();
	}
	const std::vector<const void *> &child_results(unsigned task) const
	{
		assert(task < _child_results.size());
		return _child_results[task];
	}
	std::vector<const void *> &child_results(unsigned task)
	{
		assert(task < _child_results.size());
		return _child_results[task];
	}

#ifdef QUEUE_DEBUG
	virtual unsigned depth() const
	{	return _depth; }
#endif

protected:
	// no copying
	Task_group(const Task_group &) { throw std::exception(); }
	void operator=(const Task_group &) { throw std::exception(); }

	std::vector<cl::Buffer>		_args;
	std::vector<cl::Buffer>		_spawn_args;
	std::vector<std::pair<cl_ulong *, size_t> >	_spawn_arg_bufs;
	std::vector<unsigned>		_parent_offsets;
	std::vector<std::vector<const void *> > _child_results;
	cl::Buffer	_response;
	cl::Buffer	_spawn;
	// using cl_ulong to ensure it is properly aligned (though I think
	// malloc() will make it aligned anyway ...)
	cl_ulong	*_response_buf;
	cl_uint		*_spawn_buf;
	size_t		_response_sz;

	Task_group	*_parent;
	int		_num_spawns;

#ifdef QUEUE_DEBUG
	unsigned	_depth;
#endif
};

class Root_task : public Task_group {
public:
	Root_task(cl::Context &ctx, unsigned num_tasks, size_t response_sz) :
		Task_group(ctx, 1, num_tasks, response_sz, 0, 0),
		_ctx(ctx),
		_num_tasks(num_tasks)
	{}
	virtual ~Root_task() {}

	/*
	 * for all add_arg(), convert the argument to the best OpenCL type
	 * before making the buffers in _add_arg()
	 */

	/**
	 * Add a primitive argument to the root task.
	 */
	void add_arg(cl_char x) throw (Opencl_error)
	{	_add_arg(x); }
	void add_arg(cl_uchar x) throw (Opencl_error)
	{	_add_arg(x); }

	void add_arg(cl_short x) throw (Opencl_error)
	{	_add_arg(x); }
	void add_arg(cl_ushort x) throw (Opencl_error)
	{	_add_arg(x); }

	void add_arg(cl_int x) throw (Opencl_error)
	{	_add_arg(x); }
	void add_arg(cl_uint x) throw (Opencl_error)
	{	_add_arg(x); }

	void add_arg(cl_long x) throw (Opencl_error)
	{	_add_arg(x); }
	void add_arg(cl_ulong x) throw (Opencl_error)
	{	_add_arg(x); }

	void add_arg(cl_float x) throw (Opencl_error)
	{	_add_arg(x); }
	void add_arg(cl_double x) throw (Opencl_error)
	{	_add_arg(x); }

	/**
	 * Add a generic argument/buffer to the root task.  It *should* be
	 * composed of OpenCL types only.
	 */
	void add_arg_generic(const void *val, size_t size)
	    throw (Opencl_error);

	virtual unsigned tasks() const
	{	return 1; }

	virtual unsigned child_tasks() const
	{	return _num_tasks; }

	const std::vector<size_t> &arg_sizes() const
	{	return _arg_sizes; }

protected:
	// no copying
	Root_task(const Root_task &t) : Task_group(t), _ctx(t._ctx) {}
	void operator=(const Root_task &) { throw std::exception(); }

private:
	template <typename T>
	void _add_arg(const T &t) throw (Opencl_error)
	{
		add_arg_generic(&t, sizeof(T));
	}

	std::vector<size_t>	_arg_sizes;
	cl::Context		&_ctx;
	unsigned		_num_tasks;
};

class Child_group : public Task_group {
public:
	Child_group(cl::Context &ctx, size_t response_sz,
	    const Root_task &root, Task_group &parent_group,
	    unsigned parent_offset) throw (Opencl_error);
	virtual ~Child_group() {}

	virtual unsigned tasks() const
	{	return _num_tasks; }

	virtual unsigned child_tasks() const
	{	return _num_child_tasks; }

protected:
	// no copying
	Child_group(const Child_group &t) : Task_group(t) {}
	void operator=(const Child_group &) { throw std::exception(); }

private:
	unsigned	_num_tasks;
	unsigned	_num_child_tasks;
};

#endif
