#!/usr/bin/python
# vim: set fileencoding=utf8 :

import os
import re
import sys

import matplotlib.font_manager
import matplotlib.pyplot

EVENT_KERNEL_START = 0
EVENT_KERNEL_END = 1
EVENT_REORDER_START = 2
EVENT_REORDER_END = 3
EVENT_REDUCE_START = 4
EVENT_REDUCE_END = 5
EVENT_MEM_START = 6
EVENT_MEM_END = 7
EVENT_ASYNC_START = 8
EVENT_STRING = {
	EVENT_KERNEL_START: 'start kernel invocation',
	EVENT_KERNEL_END: 'finish kernel invocation',
	EVENT_REORDER_START: 'start reordering buffers',
	EVENT_REORDER_END: 'finish reordering buffers',
	EVENT_REDUCE_START: 'start reducing tasks',
	EVENT_REDUCE_END: 'finish reducing tasks',
	EVENT_MEM_START: 'start memory transfer',
	EVENT_MEM_END: 'finish memory transfer',
	EVENT_ASYNC_START: 'start asynchronous call',
}
# invert mapping
EVENT_ID = {}
for (k, v) in EVENT_STRING.iteritems():
	EVENT_ID[v] = k

FONT_FAMILY = 'DejaVu Sans'
FONT_LEGEND = matplotlib.font_manager.FontProperties(FONT_FAMILY, size=7)
FONT_SIZE = 8 # for labels

DIMEN_SMALL = (8, 3)
DIMEN_NARROW = (8, 6)
DIMEN_WIDE = (16.5, 1) # limit seems to be ~16.54cm for letterpaper

SCALE_LIN = 0
SCALE_LOG2 = 1

GRAPH_PERIODS = 0
GRAPHPARAM = {
	GRAPH_PERIODS: {
		'xrange': (0, None),
		'yrange': (0, None),
		'xscale': SCALE_LIN,
		'yscale': SCALE_LIN,
		'xlabel': 'Iteration',
		'ylabel': 'Time (ms)',
		'legend': 'upper left',
		'dimen': DIMEN_NARROW,
		'stddev': False,
	},
}

prog = sys.argv[0]




def read_file(event_file):
	'''Read a file (object or pathname) and return a list of
	(event_id, time) tuples'''
	if not isinstance(event_file, file):
		event_file = open(event_file, 'r')
	event_re = re.compile(r'^EVENT (.*): (.*)$')
	events = []
	skipped = []
	lineno = 0
	while True:
		lineno += 1
		line = event_file.readline()
		if not len(line):
			# EOF
			break
		line = line.rstrip()
		if not len(line):
			# blank line
			continue

		m = event_re.match(line)
		if not m:
			skipped.append(lineno)
			continue
		(event_str, time) = m.groups()
		events.append((EVENT_ID[event_str], float(time)))
	if skipped:
		# print line numbers of skipped lines
		msg = '%s: skipped %d lines:' % (prog, len(skipped))
		for lineno in skipped:
			msg += ' %d' % lineno
		print >>sys.stderr, msg
	return events

def get_periods(data, start_id, end_id):
	periods = []
	start = None
	for (event, time) in data:
		if start is None:
			if event == start_id:
				start = time
		else:
			if event == end_id:
				periods.append(time - start)
				start = None
	return periods

def get_block_periods(data):
	starts = []
	ends = []
	for (event, time) in data:
		if event == EVENT_KERNEL_END:
			starts.append(time)
		elif event == EVENT_REDUCE_END:
			ends.append(time)
	assert len(starts) == len(ends)
	return [ end - start for (start, end) in zip(starts, ends) ]

def get_intervals(data, start_id, end_id):
	intervals = []
	start = None
	for (event, time) in data:
		if start is None:
			if event == start_id:
				start = time
		else:
			if event == end_id:
				intervals.append((start, time))
				start = None
	return intervals

def plot(graphtype, extraparams, formats, name, data_sets):
	'''Plot sets of data according to rules in graphtype. formats is a
	list of file extensions, name is the basename of the file.'''

	# combine GRAPHPARAM[graphtype] with extraparams
	graphparam = {}
	for (k, v) in GRAPHPARAM[graphtype].iteritems():
		graphparam[k] = v
	if extraparams:
		for (k, v) in extraparams.iteritems():
			graphparam[k] = v

	fig = matplotlib.pyplot.figure()
	dimen = graphparam['dimen']
	fig.set_size_inches(dimen[0] / 2.54, dimen[1] / 2.54)

	axes = fig.add_subplot(111)
	axes.minorticks_on();

	plots = []
	for data_set in data_sets:
		data = data_set['data']
		xdata = [ x[0] for x in data ]
		ydata = [ x[1] for x in data ]
		#if graphtype == GRAPH_TIME:
		#	ydata = [ x[1] / 1000000000. for x in data ]
		plots.append(axes.plot(xdata, ydata, data_set['format'],
		    color=data_set['color'],
		    label=data_set['label']))
		if graphparam['stddev']:
			for (x, y, e) in data:
				plots.append(axes.errorbar(x, y, yerr=e,
				    color=data_set['color']))

	# y-axis parameters
	axes.set_ylabel(graphparam['ylabel'],
	    family=FONT_FAMILY, size=FONT_SIZE)
	yrange_ = [ x for x in graphparam['yrange'] ]
	axes.set_ybound(yrange_[0], yrange_[1])
	yscale = graphparam['yscale']
	if yscale == SCALE_LOG2:
		axes.set_yscale('log', basex=2)

	# x-axis parameters
	axes.set_xlabel(graphparam['xlabel'],
	    family=FONT_FAMILY, size=FONT_SIZE)
	xrange_ = graphparam['xrange']
	axes.set_xbound(xrange_[0], xrange_[1])
	xscale = graphparam['xscale']
	if xscale == SCALE_LOG2:
		axes.set_xscale('log', basex=2)

	# set fonts of all tick labels; rescaling no longer possible
	for labels in (
	    axes.get_yticklabels(), axes.get_xticklabels()):
		for label in labels:
			label.set_family(FONT_FAMILY)
			label.set_size(FONT_SIZE)

	if graphparam.has_key('legend') and graphparam['legend'] is not None:
		axes.legend(prop=FONT_LEGEND, loc=graphparam['legend'])

	for ext in formats:
		fig.savefig('%s.%s' % (name, ext), bbox_inches='tight')

def plot_timeline(name, formats, data):
	fig = matplotlib.pyplot.figure()
	dimen = DIMEN_WIDE
	fig.set_size_inches(dimen[0] / 2.54, dimen[1] / 2.54)

	axes = fig.add_subplot(111)
	axes.minorticks_on()

	# group data by the task ID
	grouped = {}
	for (task, color, start, end) in data:
		if not grouped.has_key(task):
			grouped[task] = {
				'colors': [],
				'intervals': [],
				'label': task,
			}
		start /= 1000000.
		end   /= 1000000.
		grouped[task]['intervals'].append((start, end - start))
		grouped[task]['colors'].append(color)

	# get y-values for task and add broken bars
	tasks = sorted(grouped.keys(), reverse=True)
	for i in range(len(tasks)):
		task = tasks[i]
		group = grouped[task]
		group['tick'] = i + .5

		axes.broken_barh(group['intervals'], (i + .1, .8),
		    facecolors=group['colors'], linewidth=0)

	# x-axis parameters
	axes.set_xlim(0, None)
	axes.set_xlabel('Time (ms)',
	    family=FONT_FAMILY, size=FONT_SIZE)

	# y-axis parameters
	axes.set_ylim(0, 2)
	axes.set_yticks([grouped[task]['tick'] for task in tasks])
	axes.set_yticklabels([grouped[task]['label'] for task in tasks],
	    family=FONT_FAMILY, size=FONT_SIZE)
	# disable y tick marks
	axes.tick_params(axis='y', which='both', length=0)

	# add vertical grid
	axes.grid(True, xdata=None)

	# set fonts of all tick labels; rescaling no longer possible
	for label in axes.get_xticklabels():
		label.set_family(FONT_FAMILY)
		label.set_size(FONT_SIZE)

	for ext in formats:
		fig.savefig('%s.%s' % (name, ext),
		    bbox_inches='tight')

def timeline_data(events):
	kernels = get_intervals(events, EVENT_KERNEL_START, EVENT_KERNEL_END)
	mems = get_intervals(events, EVENT_MEM_START, EVENT_MEM_END)
	reorders = get_intervals(events, EVENT_REORDER_START,
	    EVENT_KERNEL_START)
	reduces = get_intervals(events, EVENT_REDUCE_START, EVENT_REDUCE_END)

	res = []
	for (begin, end) in kernels:
		res.append(('GPU', 'black', begin, end))
	for (begin, end) in mems:
		res.append(('GPU', 'black', begin, end))
	for (begin, end) in reorders:
		res.append(('CPU', '#808080', begin, end))
	for (begin, end) in reduces:
		res.append(('CPU', 'black', begin, end))
	return res

def graph_events(name, events):
	'''Graph events for one file.'''
	kernels = get_periods(events, EVENT_KERNEL_START, EVENT_KERNEL_END)
	reorders = get_periods(events, EVENT_REORDER_START, EVENT_REORDER_END)
	reduces = get_periods(events, EVENT_REDUCE_START, EVENT_REDUCE_END)

	directory = 'graphs'
	try:
		os.makedirs(directory)
	except OSError, e:
		if e.errno != 17:
			raise e

	kernels = [ (i, t / 1000000.) for (i, t) in enumerate(kernels) ]
	plot(GRAPH_PERIODS,
	    {
		'legend': None,
		'dimen': DIMEN_SMALL,
	    },
	    ['eps', 'png'],
	    '%s/plot-%s-period-kernel' % (directory, name),
	    [{
		'data': kernels,
		'color': '#000000',
		'format': '-',
		'label': 'Kernel',
	    }])

	reorders = [ (i, t / 1000.) for (i, t) in enumerate(reorders) ]
	reduces = [ (i, t / 1000.) for (i, t) in enumerate(reduces) ]
	plot(GRAPH_PERIODS,
	    {
		'ylabel': u'Time (µs)',
		'dimen': DIMEN_SMALL,
	    },
	    ['eps', 'png'],
	    '%s/plot-%s-period-cpu' % (directory, name),
	    [{
		'data': reorders,
		'color': '#000000',
		'format': '-',
		'label': 'Reorder',
	    }, {
		'data': reduces,
		'color': '#a0a0a0',
		'format': '-',
		'label': 'Reduce',
	    }])

	plot_timeline('%s/timeline-%s' % (directory, name), ['eps', 'png'],
	    timeline_data(events))

def graph_blocking(names, formats, event_sets):
	'''Graph events for one file.'''

	directory = 'graphs'
	try:
		os.makedirs(directory)
	except OSError, e:
		if e.errno != 17:
			raise e

	block_sets = [ get_block_periods(x) for x in event_sets ]

	# whoops, why was I averaging these?  I'll save the code anyway
	#blocks = []
	#for i in xrange(len(block_sets)):
	#	tot = 0
	#	for j in xrange(len(block_sets[i])):
	#		if not i:
	#			blocks.append(0)
	#		blocks[j] += block_sets[i][j]
	#for i in xrange(len(block_sets)):
	#	blocks[i] /= len(block_sets)

	plots = []
	for i in xrange(len(block_sets)):
		data = [ [j, t / 1000000.]
		    for (j, t) in enumerate(block_sets[i]) ]
		for j in xrange(len(data)):
			data[j][0] *= 100 / float(len(data) - 1)

		shade = (len(block_sets) - i - 1) / float(
		    len(block_sets) - 1) * .85
		shade = round(255. * shade)
		color = '#%02x%02x%02x' % (shade, shade, shade)

		plots.append({
			'color': color,
			'data': data,
			'format': formats[i],
			'label': names[i],
		})

	plot(GRAPH_PERIODS,
	    {
		'xlabel': 'Percent complete (%)',
	    },
	    ['eps', 'png'],
	    '%s/plot-block' % (directory,),
	    plots)

def main(args):
	'''The main function...'''
	events_fib = read_file('fib.log')
	events_heat = read_file('heat.log')
	events_matmul = read_file('matmul.log')
	events_nqueens = read_file('n_queens.log')
	graph_events('fib', events_fib)
	graph_events('matmul', events_matmul)
	graph_events('n_queens', events_nqueens)
	graph_events('heat', events_heat)
	graph_blocking(['fib', 'heat', 'matmul', 'n-queens'],
	    ['-+', '-s', '-*', '-o'],
	    [events_fib, events_heat, events_matmul, events_nqueens])
	return 0

sys.exit(main(sys.argv[1:]))
